/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { indentString } from '../../../mol-util/string';
import { FieldPath } from '../../../mol-io/reader/cif/schema';
function header(name, info, moldataImportPath) {
    return "/**\n * Copyright (c) 2017-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.\n *\n * Code-generated '" + name + "' schema file. " + info + "\n *\n * @author molstar/ciftools package\n */\n\nimport { Database, Column } from '" + moldataImportPath + "/db';\n\nimport Schema = Column.Schema;";
}
function footer(name) {
    return "\nexport type " + name + "_Schema = typeof " + name + "_Schema;\nexport interface " + name + "_Database extends Database<" + name + "_Schema> {};";
}
function getTypeShorthands(schema, fields) {
    var types = new Set();
    Object.keys(schema.tables).forEach(function (table) {
        if (fields && !fields[table])
            return;
        var columns = schema.tables[table].columns;
        Object.keys(columns).forEach(function (columnName) {
            if (fields && !fields[table][columnName])
                return;
            types.add(schema.tables[table].columns[columnName].type);
        });
    });
    var shorthands = [];
    types.forEach(function (type) {
        switch (type) {
            case 'str':
                shorthands.push('const str = Schema.str;');
                break;
            case 'int':
                shorthands.push('const int = Schema.int;');
                break;
            case 'float':
                shorthands.push('const float = Schema.float;');
                break;
            case 'coord':
                shorthands.push('const coord = Schema.coord;');
                break;
            case 'enum':
                shorthands.push('const Aliased = Schema.Aliased;');
                break;
            case 'matrix':
                shorthands.push('const Matrix = Schema.Matrix;');
                break;
            case 'vector':
                shorthands.push('const Vector = Schema.Vector;');
                break;
            case 'list':
                shorthands.push('const List = Schema.List;');
                break;
        }
    });
    return shorthands.join('\n');
}
function getTypeDef(c) {
    switch (c.type) {
        case 'str': return 'str';
        case 'int': return 'int';
        case 'float': return 'float';
        case 'coord': return 'coord';
        case 'enum':
            return "Aliased<'" + c.values.map(function (v) { return v.replace(/'/g, '\\\''); }).join("' | '") + "'>(" + c.subType + ")";
        case 'matrix':
            return "Matrix(" + c.rows + ", " + c.columns + ")";
        case 'vector':
            return "Vector(" + c.length + ")";
        case 'list':
            if (c.subType === 'int') {
                return "List('" + c.separator + "', x => parseInt(x, 10))";
            }
            else if (c.subType === 'float' || c.subType === 'coord') {
                return "List('" + c.separator + "', x => parseFloat(x))";
            }
            else {
                return "List('" + c.separator + "', x => x)";
            }
    }
}
var reSafePropertyName = /^[a-zA-Z_$][0-9a-zA-Z_$]*$/;
function safePropertyString(name) { return name.match(reSafePropertyName) ? name : "'" + name + "'"; }
function doc(description, spacesCount) {
    var spaces = ' '.repeat(spacesCount);
    return [
        spaces + "/**",
        ("" + indentString(description, 1, spaces + " * ")).replace(/ +\n/g, '\n'),
        spaces + " */"
    ].join('\n');
}
export function generate(name, info, schema, fields, moldataImportPath, addAliases) {
    var codeLines = [];
    if (fields) {
        Object.keys(fields).forEach(function (table) {
            if (table in schema.tables) {
                var schemaTable_1 = schema.tables[table];
                Object.keys(fields[table]).forEach(function (column) {
                    if (!(column in schemaTable_1.columns)) {
                        console.log("filter field '" + table + "." + column + "' not found in schema");
                    }
                });
            }
            else {
                console.log("filter category '" + table + "' not found in schema");
            }
        });
    }
    codeLines.push("export const " + name + "_Schema = {");
    Object.keys(schema.tables).forEach(function (table) {
        if (fields && !fields[table])
            return;
        var _a = schema.tables[table], description = _a.description, columns = _a.columns;
        if (description)
            codeLines.push(doc(description, 4));
        codeLines.push("    " + safePropertyString(table) + ": {");
        Object.keys(columns).forEach(function (columnName) {
            if (fields && !fields[table][columnName])
                return;
            var c = columns[columnName];
            var typeDef = getTypeDef(c);
            if (c.description)
                codeLines.push(doc(c.description, 8));
            codeLines.push("        " + safePropertyString(columnName) + ": " + typeDef + ",");
        });
        codeLines.push('    },');
    });
    codeLines.push('};');
    if (addAliases) {
        codeLines.push('');
        codeLines.push("export const " + name + "_Aliases = {");
        Object.keys(schema.aliases).forEach(function (path) {
            var _a = path.split('.'), table = _a[0], columnName = _a[1];
            if (fields && !fields[table])
                return;
            if (fields && !fields[table][columnName])
                return;
            var filteredAliases = new Set();
            schema.aliases[path].forEach(function (p) {
                if (!FieldPath.equal(p, path))
                    filteredAliases.add(FieldPath.canonical(p));
            });
            if (filteredAliases.size === 0)
                return;
            codeLines.push("    " + safePropertyString(path) + ": [");
            filteredAliases.forEach(function (alias) {
                codeLines.push("        '" + alias + "',");
            });
            codeLines.push('    ],');
        });
        codeLines.push('};');
    }
    return header(name, info, moldataImportPath) + "\n\n" + getTypeShorthands(schema, fields) + "\n\n" + codeLines.join('\n') + "\n" + footer(name);
}
//# sourceMappingURL=generate.js.map