"use strict";
/**
 * Copyright (c) 2018-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Viewer = exports.setDebugMode = exports.setProductionMode = exports.version = void 0;
var tslib_1 = require("tslib");
require("../../mol-util/polyfill");
var mol_plugin_1 = require("../../mol-plugin");
require("./index.html");
require("./embedded.html");
require("./favicon.ico");
var commands_1 = require("../../mol-plugin/commands");
var spec_1 = require("../../mol-plugin/spec");
var structure_1 = require("../../mol-plugin-state/actions/structure");
var config_1 = require("../../mol-plugin/config");
var cellpack_1 = require("../../extensions/cellpack");
var rcsb_1 = require("../../extensions/rcsb");
var pdbe_1 = require("../../extensions/pdbe");
var assets_1 = require("../../mol-util/assets");
var type_helpers_1 = require("../../mol-util/type-helpers");
var volume_1 = require("../../mol-plugin-state/actions/volume");
var behavior_1 = require("../../extensions/anvil/behavior");
var dnatco_1 = require("../../extensions/dnatco");
var format_1 = require("../../extensions/g3d/format");
var transforms_1 = require("../../mol-plugin-state/transforms");
var volume_representation_params_1 = require("../../mol-plugin-state/helpers/volume-representation-params");
var mp4_export_1 = require("../../extensions/mp4-export");
require('mol-plugin-ui/skin/light.scss');
var version_1 = require("../../mol-plugin/version");
Object.defineProperty(exports, "version", { enumerable: true, get: function () { return version_1.PLUGIN_VERSION; } });
var debug_1 = require("../../mol-util/debug");
Object.defineProperty(exports, "setProductionMode", { enumerable: true, get: function () { return debug_1.setProductionMode; } });
Object.defineProperty(exports, "setDebugMode", { enumerable: true, get: function () { return debug_1.setDebugMode; } });
var CustomFormats = [
    ['g3d', format_1.G3dProvider]
];
var Extensions = {
    'cellpack': spec_1.PluginSpec.Behavior(cellpack_1.CellPack),
    'dnatco-confal-pyramids': spec_1.PluginSpec.Behavior(dnatco_1.DnatcoConfalPyramids),
    'pdbe-structure-quality-report': spec_1.PluginSpec.Behavior(pdbe_1.PDBeStructureQualityReport),
    'rcsb-assembly-symmetry': spec_1.PluginSpec.Behavior(rcsb_1.RCSBAssemblySymmetry),
    'rcsb-validation-report': spec_1.PluginSpec.Behavior(rcsb_1.RCSBValidationReport),
    'anvil-membrane-orientation': spec_1.PluginSpec.Behavior(behavior_1.ANVILMembraneOrientation),
    'g3d': spec_1.PluginSpec.Behavior(format_1.G3DFormat),
    'mp4-export': spec_1.PluginSpec.Behavior(mp4_export_1.Mp4Export)
};
var DefaultViewerOptions = {
    customFormats: CustomFormats,
    extensions: type_helpers_1.ObjectKeys(Extensions),
    layoutIsExpanded: true,
    layoutShowControls: true,
    layoutShowRemoteState: true,
    layoutControlsDisplay: 'reactive',
    layoutShowSequence: true,
    layoutShowLog: true,
    layoutShowLeftPanel: true,
    disableAntialiasing: false,
    pixelScale: 1,
    enableWboit: false,
    viewportShowExpand: config_1.PluginConfig.Viewport.ShowExpand.defaultValue,
    viewportShowControls: config_1.PluginConfig.Viewport.ShowControls.defaultValue,
    viewportShowSettings: config_1.PluginConfig.Viewport.ShowSettings.defaultValue,
    viewportShowSelectionMode: config_1.PluginConfig.Viewport.ShowSelectionMode.defaultValue,
    viewportShowAnimation: config_1.PluginConfig.Viewport.ShowAnimation.defaultValue,
    pluginStateServer: config_1.PluginConfig.State.DefaultServer.defaultValue,
    volumeStreamingServer: config_1.PluginConfig.VolumeStreaming.DefaultServer.defaultValue,
    volumeStreamingDisabled: !config_1.PluginConfig.VolumeStreaming.Enabled.defaultValue,
    pdbProvider: config_1.PluginConfig.Download.DefaultPdbProvider.defaultValue,
    emdbProvider: config_1.PluginConfig.Download.DefaultEmdbProvider.defaultValue,
};
var Viewer = /** @class */ (function () {
    function Viewer(elementOrId, options) {
        if (options === void 0) { options = {}; }
        var o = tslib_1.__assign(tslib_1.__assign({}, DefaultViewerOptions), options);
        var spec = {
            actions: tslib_1.__spreadArrays(mol_plugin_1.DefaultPluginSpec.actions),
            behaviors: tslib_1.__spreadArrays(mol_plugin_1.DefaultPluginSpec.behaviors, o.extensions.map(function (e) { return Extensions[e]; })),
            animations: tslib_1.__spreadArrays(mol_plugin_1.DefaultPluginSpec.animations || []),
            customParamEditors: mol_plugin_1.DefaultPluginSpec.customParamEditors,
            customFormats: o === null || o === void 0 ? void 0 : o.customFormats,
            layout: {
                initial: {
                    isExpanded: o.layoutIsExpanded,
                    showControls: o.layoutShowControls,
                    controlsDisplay: o.layoutControlsDisplay,
                },
                controls: tslib_1.__assign(tslib_1.__assign({}, mol_plugin_1.DefaultPluginSpec.layout && mol_plugin_1.DefaultPluginSpec.layout.controls), { top: o.layoutShowSequence ? undefined : 'none', bottom: o.layoutShowLog ? undefined : 'none', left: o.layoutShowLeftPanel ? undefined : 'none' })
            },
            components: tslib_1.__assign(tslib_1.__assign({}, mol_plugin_1.DefaultPluginSpec.components), { remoteState: o.layoutShowRemoteState ? 'default' : 'none' }),
            config: [
                [config_1.PluginConfig.General.DisableAntialiasing, o.disableAntialiasing],
                [config_1.PluginConfig.General.PixelScale, o.pixelScale],
                [config_1.PluginConfig.General.EnableWboit, o.enableWboit],
                [config_1.PluginConfig.Viewport.ShowExpand, o.viewportShowExpand],
                [config_1.PluginConfig.Viewport.ShowControls, o.viewportShowControls],
                [config_1.PluginConfig.Viewport.ShowSettings, o.viewportShowSettings],
                [config_1.PluginConfig.Viewport.ShowSelectionMode, o.viewportShowSelectionMode],
                [config_1.PluginConfig.Viewport.ShowAnimation, o.viewportShowAnimation],
                [config_1.PluginConfig.State.DefaultServer, o.pluginStateServer],
                [config_1.PluginConfig.State.CurrentServer, o.pluginStateServer],
                [config_1.PluginConfig.VolumeStreaming.DefaultServer, o.volumeStreamingServer],
                [config_1.PluginConfig.VolumeStreaming.Enabled, !o.volumeStreamingDisabled],
                [config_1.PluginConfig.Download.DefaultPdbProvider, o.pdbProvider],
                [config_1.PluginConfig.Download.DefaultEmdbProvider, o.emdbProvider]
            ]
        };
        var element = typeof elementOrId === 'string'
            ? document.getElementById(elementOrId)
            : elementOrId;
        if (!element)
            throw new Error("Could not get element with id '" + elementOrId + "'");
        this.plugin = mol_plugin_1.createPlugin(element, spec);
    }
    Viewer.prototype.setRemoteSnapshot = function (id) {
        var url = this.plugin.config.get(config_1.PluginConfig.State.CurrentServer) + "/get/" + id;
        return commands_1.PluginCommands.State.Snapshots.Fetch(this.plugin, { url: url });
    };
    Viewer.prototype.loadSnapshotFromUrl = function (url, type) {
        return commands_1.PluginCommands.State.Snapshots.OpenUrl(this.plugin, { url: url, type: type });
    };
    Viewer.prototype.loadStructureFromUrl = function (url, format, isBinary, options) {
        if (format === void 0) { format = 'mmcif'; }
        if (isBinary === void 0) { isBinary = false; }
        var params = structure_1.DownloadStructure.createDefaultParams(this.plugin.state.data.root.obj, this.plugin);
        return this.plugin.runTask(this.plugin.state.data.applyAction(structure_1.DownloadStructure, {
            source: {
                name: 'url',
                params: {
                    url: assets_1.Asset.Url(url),
                    format: format,
                    isBinary: isBinary,
                    options: tslib_1.__assign(tslib_1.__assign({}, params.source.params.options), { representationParams: options === null || options === void 0 ? void 0 : options.representationParams }),
                }
            }
        }));
    };
    Viewer.prototype.loadAllModelsOrAssemblyFromUrl = function (url, format, isBinary, options) {
        if (format === void 0) { format = 'mmcif'; }
        if (isBinary === void 0) { isBinary = false; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var plugin, data, trajectory;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        plugin = this.plugin;
                        return [4 /*yield*/, plugin.builders.data.download({ url: url, isBinary: isBinary }, { state: { isGhost: true } })];
                    case 1:
                        data = _a.sent();
                        return [4 /*yield*/, plugin.builders.structure.parseTrajectory(data, format)];
                    case 2:
                        trajectory = _a.sent();
                        return [4 /*yield*/, this.plugin.builders.structure.hierarchy.applyPreset(trajectory, 'all-models', { useDefaultIfSingleModel: true, representationPresetParams: options === null || options === void 0 ? void 0 : options.representationParams })];
                    case 3:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    Viewer.prototype.loadStructureFromData = function (data, format, options) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _data, trajectory;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.plugin.builders.data.rawData({ data: data, label: options === null || options === void 0 ? void 0 : options.dataLabel })];
                    case 1:
                        _data = _a.sent();
                        return [4 /*yield*/, this.plugin.builders.structure.parseTrajectory(_data, format)];
                    case 2:
                        trajectory = _a.sent();
                        return [4 /*yield*/, this.plugin.builders.structure.hierarchy.applyPreset(trajectory, 'default')];
                    case 3:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    Viewer.prototype.loadPdb = function (pdb, options) {
        var params = structure_1.DownloadStructure.createDefaultParams(this.plugin.state.data.root.obj, this.plugin);
        var provider = this.plugin.config.get(config_1.PluginConfig.Download.DefaultPdbProvider);
        return this.plugin.runTask(this.plugin.state.data.applyAction(structure_1.DownloadStructure, {
            source: {
                name: 'pdb',
                params: {
                    provider: {
                        id: pdb,
                        server: {
                            name: provider,
                            params: structure_1.PdbDownloadProvider[provider].defaultValue
                        }
                    },
                    options: tslib_1.__assign(tslib_1.__assign({}, params.source.params.options), { representationParams: options === null || options === void 0 ? void 0 : options.representationParams }),
                }
            }
        }));
    };
    Viewer.prototype.loadPdbDev = function (pdbDev) {
        var params = structure_1.DownloadStructure.createDefaultParams(this.plugin.state.data.root.obj, this.plugin);
        return this.plugin.runTask(this.plugin.state.data.applyAction(structure_1.DownloadStructure, {
            source: {
                name: 'pdb-dev',
                params: {
                    provider: {
                        id: pdbDev,
                        encoding: 'bcif',
                    },
                    options: params.source.params.options,
                }
            }
        }));
    };
    Viewer.prototype.loadEmdb = function (emdb, options) {
        var _a;
        var provider = this.plugin.config.get(config_1.PluginConfig.Download.DefaultEmdbProvider);
        return this.plugin.runTask(this.plugin.state.data.applyAction(volume_1.DownloadDensity, {
            source: {
                name: 'pdb-emd-ds',
                params: {
                    provider: {
                        id: emdb,
                        server: provider,
                    },
                    detail: (_a = options === null || options === void 0 ? void 0 : options.detail) !== null && _a !== void 0 ? _a : 3,
                }
            }
        }));
    };
    Viewer.prototype.loadVolumeFromUrl = function (url, format, isBinary, isovalues, entryId) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var plugin;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                plugin = this.plugin;
                if (!plugin.dataFormats.get(format)) {
                    throw new Error("Unknown density format: " + format);
                }
                return [2 /*return*/, plugin.dataTransaction(function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                        var data, parsed, volume, repr, _i, isovalues_1, iso;
                        var _a;
                        return tslib_1.__generator(this, function (_b) {
                            switch (_b.label) {
                                case 0: return [4 /*yield*/, plugin.builders.data.download({ url: url, isBinary: isBinary, label: entryId }, { state: { isGhost: true } })];
                                case 1:
                                    data = _b.sent();
                                    return [4 /*yield*/, plugin.dataFormats.get(format).parse(plugin, data, { entryId: entryId })];
                                case 2:
                                    parsed = _b.sent();
                                    volume = (parsed.volume || parsed.volumes[0]);
                                    if (!(volume === null || volume === void 0 ? void 0 : volume.isOk))
                                        throw new Error('Failed to parse any volume.');
                                    repr = plugin.build().to(volume);
                                    for (_i = 0, isovalues_1 = isovalues; _i < isovalues_1.length; _i++) {
                                        iso = isovalues_1[_i];
                                        repr.apply(transforms_1.StateTransforms.Representation.VolumeRepresentation3D, volume_representation_params_1.createVolumeRepresentationParams(this.plugin, volume.data, {
                                            type: 'isosurface',
                                            typeParams: { alpha: (_a = iso.alpha) !== null && _a !== void 0 ? _a : 1, isoValue: iso.type === 'absolute' ? { kind: 'absolute', absoluteValue: iso.value } : { kind: 'relative', relativeValue: iso.value } },
                                            color: 'uniform',
                                            colorParams: { value: iso.color }
                                        }));
                                    }
                                    return [4 /*yield*/, repr.commit()];
                                case 3:
                                    _b.sent();
                                    return [2 /*return*/];
                            }
                        });
                    }); })];
            });
        });
    };
    Viewer.prototype.handleResize = function () {
        this.plugin.layout.events.updated.next();
    };
    return Viewer;
}());
exports.Viewer = Viewer;
//# sourceMappingURL=index.js.map