"use strict";
/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostprocessingPass = exports.PostprocessingParams = void 0;
var tslib_1 = require("tslib");
var util_1 = require("../../mol-gl/compute/util");
var schema_1 = require("../../mol-gl/renderable/schema");
var shader_code_1 = require("../../mol-gl/shader-code");
var mol_util_1 = require("../../mol-util");
var render_item_1 = require("../../mol-gl/webgl/render-item");
var renderable_1 = require("../../mol-gl/renderable");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var param_definition_1 = require("../../mol-util/param-definition");
var quad_vert_1 = tslib_1.__importDefault(require("../../mol-gl/shader/quad.vert"));
var postprocessing_frag_1 = tslib_1.__importDefault(require("../../mol-gl/shader/postprocessing.frag"));
var fxaa_frag_1 = tslib_1.__importDefault(require("../../mol-gl/shader/fxaa.frag"));
var stereo_1 = require("../camera/stereo");
var PostprocessingSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tColor: schema_1.TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), tPackedDepth: schema_1.TextureSpec('texture', 'depth', 'ushort', 'nearest'), uTexSize: schema_1.UniformSpec('v2'), dOrthographic: schema_1.DefineSpec('number'), uNear: schema_1.UniformSpec('f'), uFar: schema_1.UniformSpec('f'), uFogNear: schema_1.UniformSpec('f'), uFogFar: schema_1.UniformSpec('f'), uFogColor: schema_1.UniformSpec('v3'), dOcclusionEnable: schema_1.DefineSpec('boolean'), dOcclusionKernelSize: schema_1.DefineSpec('number'), uOcclusionBias: schema_1.UniformSpec('f'), uOcclusionRadius: schema_1.UniformSpec('f'), dOutlineEnable: schema_1.DefineSpec('boolean'), uOutlineScale: schema_1.UniformSpec('f'), uOutlineThreshold: schema_1.UniformSpec('f') });
var PostprocessingShaderCode = shader_code_1.ShaderCode('postprocessing', quad_vert_1.default, postprocessing_frag_1.default);
function getPostprocessingRenderable(ctx, colorTexture, depthTexture) {
    var width = colorTexture.getWidth();
    var height = colorTexture.getHeight();
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tColor: mol_util_1.ValueCell.create(colorTexture), tPackedDepth: mol_util_1.ValueCell.create(depthTexture), uTexSize: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(width, height)), dOrthographic: mol_util_1.ValueCell.create(0), uNear: mol_util_1.ValueCell.create(1), uFar: mol_util_1.ValueCell.create(10000), uFogNear: mol_util_1.ValueCell.create(10000), uFogFar: mol_util_1.ValueCell.create(10000), uFogColor: mol_util_1.ValueCell.create(linear_algebra_1.Vec3.create(1, 1, 1)), dOcclusionEnable: mol_util_1.ValueCell.create(false), dOcclusionKernelSize: mol_util_1.ValueCell.create(4), uOcclusionBias: mol_util_1.ValueCell.create(0.5), uOcclusionRadius: mol_util_1.ValueCell.create(64), dOutlineEnable: mol_util_1.ValueCell.create(false), uOutlineScale: mol_util_1.ValueCell.create(1 * ctx.pixelRatio), uOutlineThreshold: mol_util_1.ValueCell.create(0.8) });
    var schema = tslib_1.__assign({}, PostprocessingSchema);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', PostprocessingShaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
exports.PostprocessingParams = {
    occlusion: param_definition_1.ParamDefinition.MappedStatic('off', {
        on: param_definition_1.ParamDefinition.Group({
            kernelSize: param_definition_1.ParamDefinition.Numeric(4, { min: 1, max: 32, step: 1 }),
            bias: param_definition_1.ParamDefinition.Numeric(0.5, { min: 0, max: 1, step: 0.01 }),
            radius: param_definition_1.ParamDefinition.Numeric(64, { min: 0, max: 256, step: 1 }),
        }),
        off: param_definition_1.ParamDefinition.Group({})
    }, { cycle: true, description: 'Darken occluded crevices with the ambient occlusion effect' }),
    outline: param_definition_1.ParamDefinition.MappedStatic('off', {
        on: param_definition_1.ParamDefinition.Group({
            scale: param_definition_1.ParamDefinition.Numeric(1, { min: 0, max: 10, step: 1 }),
            threshold: param_definition_1.ParamDefinition.Numeric(0.8, { min: 0, max: 5, step: 0.01 }),
        }),
        off: param_definition_1.ParamDefinition.Group({})
    }, { cycle: true, description: 'Draw outline around 3D objects' }),
    antialiasing: param_definition_1.ParamDefinition.MappedStatic('on', {
        on: param_definition_1.ParamDefinition.Group({
            edgeThresholdMin: param_definition_1.ParamDefinition.Numeric(0.0312, { min: 0.0312, max: 0.0833, step: 0.0001 }, { description: 'Trims the algorithm from processing darks.' }),
            edgeThresholdMax: param_definition_1.ParamDefinition.Numeric(0.063, { min: 0.063, max: 0.333, step: 0.001 }, { description: 'The minimum amount of local contrast required to apply algorithm.' }),
            iterations: param_definition_1.ParamDefinition.Numeric(12, { min: 0, max: 32, step: 1 }, { description: 'Number of edge exploration steps.' }),
            subpixelQuality: param_definition_1.ParamDefinition.Numeric(1.00, { min: 0.00, max: 1.00, step: 0.01 }, { description: 'Choose the amount of sub-pixel aliasing removal.' }),
        }),
        off: param_definition_1.ParamDefinition.Group({})
    }, { cycle: true, description: 'Fast Approximate Anti-Aliasing (FXAA)' }),
};
var PostprocessingPass = /** @class */ (function () {
    function PostprocessingPass(webgl, drawPass) {
        this.webgl = webgl;
        this.drawPass = drawPass;
        var colorTarget = drawPass.colorTarget, depthTexture = drawPass.depthTexture;
        var width = colorTarget.getWidth();
        var height = colorTarget.getHeight();
        this.target = webgl.createRenderTarget(width, height, false);
        this.tmpTarget = webgl.createRenderTarget(width, height, false, 'uint8', 'linear');
        this.renderable = getPostprocessingRenderable(webgl, colorTarget.texture, depthTexture);
        this.fxaa = getFxaaRenderable(webgl, this.tmpTarget.texture);
    }
    PostprocessingPass.isEnabled = function (props) {
        return props.occlusion.name === 'on' || props.outline.name === 'on' || props.antialiasing.name === 'on';
    };
    PostprocessingPass.prototype.syncSize = function () {
        var width = this.drawPass.colorTarget.getWidth();
        var height = this.drawPass.colorTarget.getHeight();
        var _a = this.renderable.values.uTexSize.ref.value, w = _a[0], h = _a[1];
        if (width !== w || height !== h) {
            this.target.setSize(width, height);
            this.tmpTarget.setSize(width, height);
            mol_util_1.ValueCell.update(this.renderable.values.uTexSize, linear_algebra_1.Vec2.set(this.renderable.values.uTexSize.ref.value, width, height));
            mol_util_1.ValueCell.update(this.fxaa.values.uTexSizeInv, linear_algebra_1.Vec2.set(this.fxaa.values.uTexSizeInv.ref.value, 1 / width, 1 / height));
        }
    };
    PostprocessingPass.prototype.updateState = function (camera) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        state.enable(gl.SCISSOR_TEST);
        state.disable(gl.BLEND);
        state.disable(gl.DEPTH_TEST);
        state.depthMask(false);
        var _b = camera.viewport, x = _b.x, y = _b.y, width = _b.width, height = _b.height;
        gl.viewport(x, y, width, height);
        gl.scissor(x, y, width, height);
        state.clearColor(0, 0, 0, 1);
        gl.clear(gl.COLOR_BUFFER_BIT);
    };
    PostprocessingPass.prototype._renderPostprocessing = function (camera, toDrawingBuffer, props) {
        var values = this.renderable.values;
        mol_util_1.ValueCell.updateIfChanged(values.uFar, camera.far);
        mol_util_1.ValueCell.updateIfChanged(values.uNear, camera.near);
        mol_util_1.ValueCell.updateIfChanged(values.uFogFar, camera.fogFar);
        mol_util_1.ValueCell.updateIfChanged(values.uFogNear, camera.fogNear);
        var needsUpdate = false;
        var orthographic = camera.state.mode === 'orthographic' ? 1 : 0;
        if (values.dOrthographic.ref.value !== orthographic)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dOrthographic, orthographic);
        var occlusion = props.occlusion.name === 'on';
        if (values.dOcclusionEnable.ref.value !== occlusion)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(this.renderable.values.dOcclusionEnable, occlusion);
        if (props.occlusion.name === 'on') {
            var kernelSize = props.occlusion.params.kernelSize;
            if (values.dOcclusionKernelSize.ref.value !== kernelSize)
                needsUpdate = true;
            mol_util_1.ValueCell.updateIfChanged(values.dOcclusionKernelSize, kernelSize);
            mol_util_1.ValueCell.updateIfChanged(values.uOcclusionBias, props.occlusion.params.bias);
            mol_util_1.ValueCell.updateIfChanged(values.uOcclusionRadius, props.occlusion.params.radius);
        }
        var outline = props.outline.name === 'on';
        if (values.dOutlineEnable.ref.value !== outline)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dOutlineEnable, outline);
        if (props.outline.name === 'on') {
            mol_util_1.ValueCell.updateIfChanged(values.uOutlineScale, props.outline.params.scale * this.webgl.pixelRatio);
            mol_util_1.ValueCell.updateIfChanged(values.uOutlineThreshold, props.outline.params.threshold);
        }
        if (needsUpdate) {
            this.renderable.update();
        }
        if (props.antialiasing.name === 'on') {
            this.tmpTarget.bind();
        }
        else if (toDrawingBuffer) {
            this.webgl.unbindFramebuffer();
        }
        else {
            this.target.bind();
        }
        this.updateState(camera);
        this.renderable.render();
    };
    PostprocessingPass.prototype._renderFxaa = function (camera, toDrawingBuffer, props) {
        if (props.antialiasing.name === 'off')
            return;
        var values = this.fxaa.values;
        var needsUpdate = false;
        var input = (props.occlusion.name === 'on' || props.outline.name === 'on')
            ? this.tmpTarget.texture : this.drawPass.colorTarget.texture;
        if (values.tColor.ref.value !== input) {
            mol_util_1.ValueCell.update(this.fxaa.values.tColor, input);
            needsUpdate = true;
        }
        var _a = props.antialiasing.params, edgeThresholdMin = _a.edgeThresholdMin, edgeThresholdMax = _a.edgeThresholdMax, iterations = _a.iterations, subpixelQuality = _a.subpixelQuality;
        if (values.dEdgeThresholdMin.ref.value !== edgeThresholdMin)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dEdgeThresholdMin, edgeThresholdMin);
        if (values.dEdgeThresholdMax.ref.value !== edgeThresholdMax)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dEdgeThresholdMax, edgeThresholdMax);
        if (values.dIterations.ref.value !== iterations)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dIterations, iterations);
        if (values.dSubpixelQuality.ref.value !== subpixelQuality)
            needsUpdate = true;
        mol_util_1.ValueCell.updateIfChanged(values.dSubpixelQuality, subpixelQuality);
        if (needsUpdate) {
            this.fxaa.update();
        }
        if (toDrawingBuffer) {
            this.webgl.unbindFramebuffer();
        }
        else {
            this.target.bind();
        }
        this.updateState(camera);
        this.fxaa.render();
    };
    PostprocessingPass.prototype._render = function (camera, toDrawingBuffer, props) {
        if (props.occlusion.name === 'on' || props.outline.name === 'on' || props.antialiasing.name === 'off') {
            this._renderPostprocessing(camera, toDrawingBuffer, props);
        }
        if (props.antialiasing.name === 'on') {
            this._renderFxaa(camera, toDrawingBuffer, props);
        }
    };
    PostprocessingPass.prototype.render = function (camera, toDrawingBuffer, props) {
        if (stereo_1.StereoCamera.is(camera)) {
            this._render(camera.left, toDrawingBuffer, props);
            this._render(camera.right, toDrawingBuffer, props);
        }
        else {
            this._render(camera, toDrawingBuffer, props);
        }
    };
    return PostprocessingPass;
}());
exports.PostprocessingPass = PostprocessingPass;
//
var FxaaSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tColor: schema_1.TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), uTexSizeInv: schema_1.UniformSpec('v2'), dEdgeThresholdMin: schema_1.DefineSpec('number'), dEdgeThresholdMax: schema_1.DefineSpec('number'), dIterations: schema_1.DefineSpec('number'), dSubpixelQuality: schema_1.DefineSpec('number') });
var FxaaShaderCode = shader_code_1.ShaderCode('fxaa', quad_vert_1.default, fxaa_frag_1.default);
function getFxaaRenderable(ctx, colorTexture) {
    var width = colorTexture.getWidth();
    var height = colorTexture.getHeight();
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tColor: mol_util_1.ValueCell.create(colorTexture), uTexSizeInv: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(1 / width, 1 / height)), dEdgeThresholdMin: mol_util_1.ValueCell.create(0.0312), dEdgeThresholdMax: mol_util_1.ValueCell.create(0.125), dIterations: mol_util_1.ValueCell.create(12), dSubpixelQuality: mol_util_1.ValueCell.create(0.75) });
    var schema = tslib_1.__assign({}, FxaaSchema);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', FxaaShaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
//# sourceMappingURL=postprocessing.js.map