"use strict";
/**
 * Copyright (c) 2019-2021 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createHistogramPyramid = void 0;
var tslib_1 = require("tslib");
var renderable_1 = require("../../renderable");
var render_item_1 = require("../../webgl/render-item");
var schema_1 = require("../../renderable/schema");
var shader_code_1 = require("../../../mol-gl/shader-code");
var mol_util_1 = require("../../../mol-util");
var util_1 = require("../util");
var sum_1 = require("./sum");
var misc_1 = require("../../../mol-math/misc");
var quad_vert_1 = tslib_1.__importDefault(require("../../../mol-gl/shader/quad.vert"));
var reduction_frag_1 = tslib_1.__importDefault(require("../../../mol-gl/shader/histogram-pyramid/reduction.frag"));
var HistopyramidReductionSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tPreviousLevel: schema_1.TextureSpec('texture', 'rgba', 'float', 'nearest'), uSize: schema_1.UniformSpec('f'), uTexSize: schema_1.UniformSpec('f'), uFirst: schema_1.UniformSpec('b') });
var HistogramPyramidName = 'histogram-pyramid';
function getHistopyramidReductionRenderable(ctx, initialTexture) {
    if (ctx.namedComputeRenderables[HistogramPyramidName]) {
        var v = ctx.namedComputeRenderables[HistogramPyramidName].values;
        mol_util_1.ValueCell.update(v.tPreviousLevel, initialTexture);
        ctx.namedComputeRenderables[HistogramPyramidName].update();
    }
    else {
        ctx.namedComputeRenderables[HistogramPyramidName] = createHistopyramidReductionRenderable(ctx, initialTexture);
    }
    return ctx.namedComputeRenderables[HistogramPyramidName];
}
function createHistopyramidReductionRenderable(ctx, initialTexture) {
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tPreviousLevel: mol_util_1.ValueCell.create(initialTexture), uSize: mol_util_1.ValueCell.create(0), uTexSize: mol_util_1.ValueCell.create(0), uFirst: mol_util_1.ValueCell.create(true) });
    var schema = tslib_1.__assign({}, HistopyramidReductionSchema);
    var shaderCode = shader_code_1.ShaderCode('reduction', quad_vert_1.default, reduction_frag_1.default);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
var LevelTexturesFramebuffers = [];
function getLevelTextureFramebuffer(ctx, level) {
    var textureFramebuffer = LevelTexturesFramebuffers[level];
    var size = Math.pow(2, level);
    if (textureFramebuffer === undefined) {
        var texture = getTexture("level" + level, ctx, 'image-float32', 'rgba', 'float', 'nearest');
        var framebuffer = getFramebuffer("level" + level, ctx);
        texture.attachFramebuffer(framebuffer, 0);
        textureFramebuffer = { texture: texture, framebuffer: framebuffer };
        textureFramebuffer.texture.define(size, size);
        LevelTexturesFramebuffers[level] = textureFramebuffer;
    }
    return textureFramebuffer;
}
function setRenderingDefaults(ctx) {
    var gl = ctx.gl, state = ctx.state;
    state.disable(gl.CULL_FACE);
    state.disable(gl.BLEND);
    state.disable(gl.DEPTH_TEST);
    state.enable(gl.SCISSOR_TEST);
    state.depthMask(false);
    state.colorMask(true, true, true, true);
    state.clearColor(0, 0, 0, 0);
}
function getFramebuffer(name, webgl) {
    var _name = HistogramPyramidName + "-" + name;
    if (!webgl.namedFramebuffers[_name]) {
        webgl.namedFramebuffers[_name] = webgl.resources.framebuffer();
    }
    return webgl.namedFramebuffers[_name];
}
function getTexture(name, webgl, kind, format, type, filter) {
    var _name = HistogramPyramidName + "-" + name;
    if (!webgl.namedTextures[_name]) {
        webgl.namedTextures[_name] = webgl.resources.texture(kind, format, type, filter);
    }
    return webgl.namedTextures[_name];
}
function createHistogramPyramid(ctx, inputTexture, scale, gridTexDim) {
    var gl = ctx.gl;
    var w = inputTexture.getWidth();
    var h = inputTexture.getHeight();
    // printTexture(ctx, inputTexture, 2)
    if (w !== h || !misc_1.isPowerOfTwo(w)) {
        throw new Error('inputTexture must be of square power-of-two size');
    }
    // This part set the levels
    var levels = Math.ceil(Math.log(w) / Math.log(2));
    var maxSize = Math.pow(2, levels);
    // console.log('levels', levels, 'maxSize', maxSize, 'input', w);
    var pyramidTexture = getTexture('pyramid', ctx, 'image-float32', 'rgba', 'float', 'nearest');
    pyramidTexture.define(maxSize, maxSize);
    var framebuffer = getFramebuffer('pyramid', ctx);
    pyramidTexture.attachFramebuffer(framebuffer, 0);
    gl.viewport(0, 0, maxSize, maxSize);
    gl.clear(gl.COLOR_BUFFER_BIT);
    var levelTexturesFramebuffers = [];
    for (var i = 0; i < levels; ++i)
        levelTexturesFramebuffers.push(getLevelTextureFramebuffer(ctx, i));
    var renderable = getHistopyramidReductionRenderable(ctx, inputTexture);
    ctx.state.currentRenderItemId = -1;
    setRenderingDefaults(ctx);
    var offset = 0;
    for (var i = 0; i < levels; i++) {
        var currLevel = levels - 1 - i;
        var tf = levelTexturesFramebuffers[currLevel];
        tf.framebuffer.bind();
        var size = Math.pow(2, currLevel);
        // console.log('size', size, 'draw-level', currLevel, 'read-level', levels - i)
        mol_util_1.ValueCell.update(renderable.values.uSize, Math.pow(2, i + 1) / maxSize);
        mol_util_1.ValueCell.update(renderable.values.uTexSize, size);
        mol_util_1.ValueCell.updateIfChanged(renderable.values.uFirst, i === 0);
        if (i > 0) {
            mol_util_1.ValueCell.update(renderable.values.tPreviousLevel, levelTexturesFramebuffers[levels - i].texture);
            renderable.update();
        }
        ctx.state.currentRenderItemId = -1;
        gl.viewport(0, 0, size, size);
        gl.scissor(0, 0, size, size);
        gl.clear(gl.COLOR_BUFFER_BIT);
        gl.scissor(0, 0, gridTexDim[0], gridTexDim[1]);
        renderable.render();
        pyramidTexture.bind(0);
        gl.copyTexSubImage2D(gl.TEXTURE_2D, 0, offset, 0, 0, 0, size, size);
        pyramidTexture.unbind(0);
        offset += size;
    }
    gl.finish();
    // printTexture(ctx, pyramidTexture, 2)
    //
    var finalCount = sum_1.getHistopyramidSum(ctx, levelTexturesFramebuffers[0].texture);
    var height = Math.ceil(finalCount / Math.pow(2, levels));
    // const scale = Vec2.create(maxSize / inputTexture.width, maxSize / inputTexture.height);
    // console.log('height', height, 'finalCount', finalCount, 'scale', scale);
    return {
        pyramidTex: pyramidTexture,
        count: finalCount,
        height: height,
        levels: levels,
        scale: scale
    };
}
exports.createHistogramPyramid = createHistogramPyramid;
//# sourceMappingURL=reduction.js.map