"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.PrincipalAxes = void 0;
var tslib_1 = require("tslib");
var matrix_1 = tslib_1.__importDefault(require("./matrix"));
var _3d_1 = require("../3d");
var svd_1 = require("./svd");
var geometry_1 = require("../../geometry");
var PrincipalAxes;
(function (PrincipalAxes) {
    function ofPositions(positions) {
        var momentsAxes = calculateMomentsAxes(positions);
        var boxAxes = calculateBoxAxes(positions, momentsAxes);
        return { momentsAxes: momentsAxes, boxAxes: boxAxes };
    }
    PrincipalAxes.ofPositions = ofPositions;
    function calculateMomentsAxes(positions) {
        if (positions.length === 3) {
            return geometry_1.Axes3D.create(_3d_1.Vec3.fromArray(_3d_1.Vec3(), positions, 0), _3d_1.Vec3.create(1, 0, 0), _3d_1.Vec3.create(0, 1, 0), _3d_1.Vec3.create(0, 1, 0));
        }
        var points = matrix_1.default.fromArray(positions, 3, positions.length / 3);
        var n = points.rows;
        var n3 = n / 3;
        var A = matrix_1.default.create(3, 3);
        var W = matrix_1.default.create(1, 3);
        var U = matrix_1.default.create(3, 3);
        var V = matrix_1.default.create(3, 3);
        // calculate
        var mean = matrix_1.default.meanRows(points);
        var pointsM = matrix_1.default.subRows(matrix_1.default.clone(points), mean);
        var pointsT = matrix_1.default.transpose(matrix_1.default.create(n, 3), pointsM);
        matrix_1.default.multiplyABt(A, pointsT, pointsT);
        svd_1.svd(A, W, U, V);
        // origin
        var origin = _3d_1.Vec3.create(mean[0], mean[1], mean[2]);
        // directions
        var dirA = _3d_1.Vec3.create(U.data[0], U.data[3], U.data[6]);
        var dirB = _3d_1.Vec3.create(U.data[1], U.data[4], U.data[7]);
        var dirC = _3d_1.Vec3.create(U.data[2], U.data[5], U.data[8]);
        _3d_1.Vec3.scale(dirA, dirA, Math.sqrt(W.data[0] / n3));
        _3d_1.Vec3.scale(dirB, dirB, Math.sqrt(W.data[1] / n3));
        _3d_1.Vec3.scale(dirC, dirC, Math.sqrt(W.data[2] / n3));
        return geometry_1.Axes3D.create(origin, dirA, dirB, dirC);
    }
    PrincipalAxes.calculateMomentsAxes = calculateMomentsAxes;
    var tmpBoxVec = _3d_1.Vec3();
    var tmpBoxVecA = _3d_1.Vec3();
    var tmpBoxVecB = _3d_1.Vec3();
    var tmpBoxVecC = _3d_1.Vec3();
    /**
     * Get the scale/length for each dimension for a box around the axes
     * to enclose the given positions
     */
    function calculateBoxAxes(positions, momentsAxes) {
        if (positions.length === 3) {
            return geometry_1.Axes3D.clone(momentsAxes);
        }
        var d1a = -Infinity;
        var d1b = -Infinity;
        var d2a = -Infinity;
        var d2b = -Infinity;
        var d3a = -Infinity;
        var d3b = -Infinity;
        var p = _3d_1.Vec3();
        var t = _3d_1.Vec3();
        var center = momentsAxes.origin;
        var normVecA = _3d_1.Vec3.normalize(tmpBoxVecA, momentsAxes.dirA);
        var normVecB = _3d_1.Vec3.normalize(tmpBoxVecB, momentsAxes.dirB);
        var normVecC = _3d_1.Vec3.normalize(tmpBoxVecC, momentsAxes.dirC);
        for (var i = 0, il = positions.length; i < il; i += 3) {
            _3d_1.Vec3.projectPointOnVector(p, _3d_1.Vec3.fromArray(p, positions, i), normVecA, center);
            var dp1 = _3d_1.Vec3.dot(normVecA, _3d_1.Vec3.normalize(t, _3d_1.Vec3.sub(t, p, center)));
            var dt1 = _3d_1.Vec3.distance(p, center);
            if (dp1 > 0) {
                if (dt1 > d1a)
                    d1a = dt1;
            }
            else {
                if (dt1 > d1b)
                    d1b = dt1;
            }
            _3d_1.Vec3.projectPointOnVector(p, _3d_1.Vec3.fromArray(p, positions, i), normVecB, center);
            var dp2 = _3d_1.Vec3.dot(normVecB, _3d_1.Vec3.normalize(t, _3d_1.Vec3.sub(t, p, center)));
            var dt2 = _3d_1.Vec3.distance(p, center);
            if (dp2 > 0) {
                if (dt2 > d2a)
                    d2a = dt2;
            }
            else {
                if (dt2 > d2b)
                    d2b = dt2;
            }
            _3d_1.Vec3.projectPointOnVector(p, _3d_1.Vec3.fromArray(p, positions, i), normVecC, center);
            var dp3 = _3d_1.Vec3.dot(normVecC, _3d_1.Vec3.normalize(t, _3d_1.Vec3.sub(t, p, center)));
            var dt3 = _3d_1.Vec3.distance(p, center);
            if (dp3 > 0) {
                if (dt3 > d3a)
                    d3a = dt3;
            }
            else {
                if (dt3 > d3b)
                    d3b = dt3;
            }
        }
        var dirA = _3d_1.Vec3.setMagnitude(_3d_1.Vec3(), normVecA, (d1a + d1b) / 2);
        var dirB = _3d_1.Vec3.setMagnitude(_3d_1.Vec3(), normVecB, (d2a + d2b) / 2);
        var dirC = _3d_1.Vec3.setMagnitude(_3d_1.Vec3(), normVecC, (d3a + d3b) / 2);
        var origin = _3d_1.Vec3();
        var addCornerHelper = function (d1, d2, d3) {
            _3d_1.Vec3.copy(tmpBoxVec, center);
            _3d_1.Vec3.scaleAndAdd(tmpBoxVec, tmpBoxVec, normVecA, d1);
            _3d_1.Vec3.scaleAndAdd(tmpBoxVec, tmpBoxVec, normVecB, d2);
            _3d_1.Vec3.scaleAndAdd(tmpBoxVec, tmpBoxVec, normVecC, d3);
            _3d_1.Vec3.add(origin, origin, tmpBoxVec);
        };
        addCornerHelper(d1a, d2a, d3a);
        addCornerHelper(d1a, d2a, -d3b);
        addCornerHelper(d1a, -d2b, -d3b);
        addCornerHelper(d1a, -d2b, d3a);
        addCornerHelper(-d1b, -d2b, -d3b);
        addCornerHelper(-d1b, -d2b, d3a);
        addCornerHelper(-d1b, d2a, d3a);
        addCornerHelper(-d1b, d2a, -d3b);
        _3d_1.Vec3.scale(origin, origin, 1 / 8);
        return geometry_1.Axes3D.create(origin, dirA, dirB, dirC);
    }
    PrincipalAxes.calculateBoxAxes = calculateBoxAxes;
})(PrincipalAxes || (PrincipalAxes = {}));
exports.PrincipalAxes = PrincipalAxes;
//# sourceMappingURL=principal-axes.js.map