"use strict";
/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.IntraUnitBondLineVisual = exports.IntraUnitBondLineParams = void 0;
var tslib_1 = require("tslib");
var param_definition_1 = require("../../../mol-util/param-definition");
var structure_1 = require("../../../mol-model/structure");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var mol_util_1 = require("../../../mol-util");
var link_1 = require("./util/link");
var units_visual_1 = require("../units-visual");
var types_1 = require("../../../mol-model/structure/model/types");
var bond_1 = require("./util/bond");
var geometry_1 = require("../../../mol-math/geometry");
var lines_1 = require("../../../mol-geo/geometry/lines/lines");
var graph_1 = require("../../../mol-math/graph");
// avoiding namespace lookup improved performance in Chrome (Aug 2020)
var isBondType = types_1.BondType.is;
function createIntraUnitBondLines(ctx, unit, structure, theme, props, lines) {
    if (!structure_1.Unit.isAtomic(unit))
        return lines_1.Lines.createEmpty(lines);
    var location = structure_1.StructureElement.Location.create(structure, unit);
    var elements = unit.elements;
    var bonds = unit.bonds;
    var edgeCount = bonds.edgeCount, a = bonds.a, b = bonds.b, edgeProps = bonds.edgeProps, offset = bonds.offset;
    var _order = edgeProps.order, _flags = edgeProps.flags;
    var sizeFactor = props.sizeFactor;
    if (!edgeCount)
        return lines_1.Lines.createEmpty(lines);
    var vRef = linear_algebra_1.Vec3();
    var pos = unit.conformation.invariantPosition;
    var builderProps = {
        linkCount: edgeCount * 2,
        referencePosition: function (edgeIndex) {
            var _a, _b;
            var aI = a[edgeIndex], bI = b[edgeIndex];
            if (aI > bI)
                _a = [bI, aI], aI = _a[0], bI = _a[1];
            if (offset[aI + 1] - offset[aI] === 1)
                _b = [bI, aI], aI = _b[0], bI = _b[1];
            // TODO prefer reference atoms within rings
            for (var i = offset[aI], il = offset[aI + 1]; i < il; ++i) {
                var _bI = b[i];
                if (_bI !== bI && _bI !== aI)
                    return pos(elements[_bI], vRef);
            }
            for (var i = offset[bI], il = offset[bI + 1]; i < il; ++i) {
                var _aI = a[i];
                if (_aI !== aI && _aI !== bI)
                    return pos(elements[_aI], vRef);
            }
            return null;
        },
        position: function (posA, posB, edgeIndex) {
            pos(elements[a[edgeIndex]], posA);
            pos(elements[b[edgeIndex]], posB);
        },
        style: function (edgeIndex) {
            var o = _order[edgeIndex];
            var f = _flags[edgeIndex];
            if (isBondType(f, 2 /* MetallicCoordination */) || isBondType(f, 4 /* HydrogenBond */)) {
                // show metall coordinations and hydrogen bonds with dashed cylinders
                return 1 /* Dashed */;
            }
            else if (o === 2) {
                return 2 /* Double */;
            }
            else if (o === 3) {
                return 3 /* Triple */;
            }
            else {
                return 0 /* Solid */;
            }
        },
        radius: function (edgeIndex) {
            location.element = elements[a[edgeIndex]];
            var sizeA = theme.size.size(location);
            location.element = elements[b[edgeIndex]];
            var sizeB = theme.size.size(location);
            return Math.min(sizeA, sizeB) * sizeFactor;
        },
        ignore: bond_1.makeIntraBondIgnoreTest(unit, props)
    };
    var l = link_1.createLinkLines(ctx, builderProps, props, lines);
    var sphere = geometry_1.Sphere3D.expand(geometry_1.Sphere3D(), unit.boundary.sphere, 1 * sizeFactor);
    l.setBoundingSphere(sphere);
    return l;
}
exports.IntraUnitBondLineParams = tslib_1.__assign(tslib_1.__assign({}, units_visual_1.UnitsLinesParams), bond_1.BondLineParams);
function IntraUnitBondLineVisual(materialId) {
    return units_visual_1.UnitsLinesVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.IntraUnitBondLineParams),
        createGeometry: createIntraUnitBondLines,
        createLocationIterator: bond_1.BondIterator.fromGroup,
        getLoci: bond_1.getIntraBondLoci,
        eachLocation: bond_1.eachIntraBond,
        setUpdateState: function (state, newProps, currentProps, newTheme, currentTheme, newStructureGroup, currentStructureGroup) {
            state.createGeometry = (newProps.sizeFactor !== currentProps.sizeFactor ||
                newProps.linkScale !== currentProps.linkScale ||
                newProps.linkSpacing !== currentProps.linkSpacing ||
                newProps.ignoreHydrogens !== currentProps.ignoreHydrogens ||
                !mol_util_1.arrayEqual(newProps.includeTypes, currentProps.includeTypes) ||
                !mol_util_1.arrayEqual(newProps.excludeTypes, currentProps.excludeTypes));
            var newUnit = newStructureGroup.group.units[0];
            var currentUnit = currentStructureGroup.group.units[0];
            if (structure_1.Unit.isAtomic(newUnit) && structure_1.Unit.isAtomic(currentUnit)) {
                if (!graph_1.IntAdjacencyGraph.areEqual(newUnit.bonds, currentUnit.bonds)) {
                    state.createGeometry = true;
                    state.updateTransform = true;
                    state.updateColor = true;
                    state.updateSize = true;
                }
            }
        }
    }, materialId);
}
exports.IntraUnitBondLineVisual = IntraUnitBondLineVisual;
//# sourceMappingURL=bond-intra-unit-line.js.map