"use strict";
/**
 * Copyright (c) 2018-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createLinkLines = exports.createLinkCylinderMesh = exports.calculateShiftDir = exports.DefaultLinkLineProps = exports.LinkLineParams = exports.DefaultLinkCylinderProps = exports.LinkCylinderParams = void 0;
var linear_algebra_1 = require("../../../../mol-math/linear-algebra");
var param_definition_1 = require("../../../../mol-util/param-definition");
var mesh_1 = require("../../../../mol-geo/geometry/mesh/mesh");
var mesh_builder_1 = require("../../../../mol-geo/geometry/mesh/mesh-builder");
var cylinder_1 = require("../../../../mol-geo/geometry/mesh/builder/cylinder");
var base_1 = require("../../../../mol-geo/geometry/base");
var lines_1 = require("../../../../mol-geo/geometry/lines/lines");
var lines_builder_1 = require("../../../../mol-geo/geometry/lines/lines-builder");
exports.LinkCylinderParams = {
    linkScale: param_definition_1.ParamDefinition.Numeric(0.4, { min: 0, max: 1, step: 0.1 }),
    linkSpacing: param_definition_1.ParamDefinition.Numeric(1, { min: 0, max: 2, step: 0.01 }),
    linkCap: param_definition_1.ParamDefinition.Boolean(false),
    radialSegments: param_definition_1.ParamDefinition.Numeric(16, { min: 2, max: 56, step: 2 }, base_1.BaseGeometry.CustomQualityParamInfo),
};
exports.DefaultLinkCylinderProps = param_definition_1.ParamDefinition.getDefaultValues(exports.LinkCylinderParams);
exports.LinkLineParams = {
    linkScale: param_definition_1.ParamDefinition.Numeric(0.5, { min: 0, max: 1, step: 0.1 }),
    linkSpacing: param_definition_1.ParamDefinition.Numeric(0.1, { min: 0, max: 2, step: 0.01 }),
};
exports.DefaultLinkLineProps = param_definition_1.ParamDefinition.getDefaultValues(exports.LinkLineParams);
var tmpV12 = linear_algebra_1.Vec3();
var tmpShiftV12 = linear_algebra_1.Vec3();
var tmpShiftV13 = linear_algebra_1.Vec3();
var up = linear_algebra_1.Vec3.create(0, 1, 0);
/** Calculate 'shift' direction that is perpendiculat to v1 - v2 and goes through v3 */
function calculateShiftDir(out, v1, v2, v3) {
    linear_algebra_1.Vec3.normalize(tmpShiftV12, linear_algebra_1.Vec3.sub(tmpShiftV12, v1, v2));
    if (v3 !== null) {
        linear_algebra_1.Vec3.sub(tmpShiftV13, v1, v3);
    }
    else {
        linear_algebra_1.Vec3.copy(tmpShiftV13, v1); // no reference point, use v1
    }
    linear_algebra_1.Vec3.normalize(tmpShiftV13, tmpShiftV13);
    // ensure v13 and v12 are not colinear
    var dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
    if (1 - Math.abs(dp) < 1e-5) {
        linear_algebra_1.Vec3.set(tmpShiftV13, 1, 0, 0);
        dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
        if (1 - Math.abs(dp) < 1e-5) {
            linear_algebra_1.Vec3.set(tmpShiftV13, 0, 1, 0);
            dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
        }
    }
    linear_algebra_1.Vec3.setMagnitude(tmpShiftV12, tmpShiftV12, dp);
    linear_algebra_1.Vec3.sub(tmpShiftV13, tmpShiftV13, tmpShiftV12);
    return linear_algebra_1.Vec3.normalize(out, tmpShiftV13);
}
exports.calculateShiftDir = calculateShiftDir;
// avoiding namespace lookup improved performance in Chrome (Aug 2020)
var v3scale = linear_algebra_1.Vec3.scale;
var v3add = linear_algebra_1.Vec3.add;
var v3sub = linear_algebra_1.Vec3.sub;
var v3setMagnitude = linear_algebra_1.Vec3.setMagnitude;
var v3dot = linear_algebra_1.Vec3.dot;
/**
 * Each edge is included twice to allow for coloring/picking
 * the half closer to the first vertex, i.e. vertex a.
 */
function createLinkCylinderMesh(ctx, linkBuilder, props, mesh) {
    var linkCount = linkBuilder.linkCount, referencePosition = linkBuilder.referencePosition, position = linkBuilder.position, style = linkBuilder.style, radius = linkBuilder.radius, ignore = linkBuilder.ignore;
    if (!linkCount)
        return mesh_1.Mesh.createEmpty(mesh);
    var linkScale = props.linkScale, linkSpacing = props.linkSpacing, radialSegments = props.radialSegments, linkCap = props.linkCap;
    var vertexCountEstimate = radialSegments * 2 * linkCount * 2;
    var builderState = mesh_builder_1.MeshBuilder.createState(vertexCountEstimate, vertexCountEstimate / 4, mesh);
    var va = linear_algebra_1.Vec3();
    var vb = linear_algebra_1.Vec3();
    var vShift = linear_algebra_1.Vec3();
    var cylinderProps = {
        radiusTop: 1,
        radiusBottom: 1,
        radialSegments: radialSegments,
        topCap: linkCap,
        bottomCap: linkCap
    };
    for (var edgeIndex = 0, _eI = linkCount; edgeIndex < _eI; ++edgeIndex) {
        if (ignore && ignore(edgeIndex))
            continue;
        position(va, vb, edgeIndex);
        var linkRadius = radius(edgeIndex);
        var linkStyle = style ? style(edgeIndex) : 0 /* Solid */;
        builderState.currentGroup = edgeIndex;
        if (linkStyle === 0 /* Solid */) {
            cylinderProps.radiusTop = cylinderProps.radiusBottom = linkRadius;
            cylinderProps.topCap = cylinderProps.bottomCap = linkCap;
            cylinder_1.addCylinder(builderState, va, vb, 0.5, cylinderProps);
        }
        else if (linkStyle === 1 /* Dashed */) {
            cylinderProps.radiusTop = cylinderProps.radiusBottom = linkRadius / 3;
            cylinderProps.topCap = cylinderProps.bottomCap = true;
            cylinder_1.addFixedCountDashedCylinder(builderState, va, vb, 0.5, 7, cylinderProps);
        }
        else if (linkStyle === 2 /* Double */ || linkStyle === 3 /* Triple */) {
            var order = linkStyle === 2 /* Double */ ? 2 : 3;
            var multiRadius = linkRadius * (linkScale / (0.5 * order));
            var absOffset = (linkRadius - multiRadius) * linkSpacing;
            calculateShiftDir(vShift, va, vb, referencePosition ? referencePosition(edgeIndex) : null);
            v3setMagnitude(vShift, vShift, absOffset);
            cylinderProps.radiusTop = cylinderProps.radiusBottom = multiRadius;
            cylinderProps.topCap = cylinderProps.bottomCap = linkCap;
            if (order === 3)
                cylinder_1.addCylinder(builderState, va, vb, 0.5, cylinderProps);
            cylinder_1.addDoubleCylinder(builderState, va, vb, 0.5, vShift, cylinderProps);
        }
        else if (linkStyle === 4 /* Disk */) {
            v3scale(tmpV12, v3sub(tmpV12, vb, va), 0.475);
            v3add(va, va, tmpV12);
            v3sub(vb, vb, tmpV12);
            cylinderProps.radiusTop = cylinderProps.radiusBottom = linkRadius;
            if (v3dot(tmpV12, up) > 0) {
                cylinderProps.topCap = false;
                cylinderProps.bottomCap = linkCap;
            }
            else {
                cylinderProps.topCap = linkCap;
                cylinderProps.bottomCap = false;
            }
            cylinder_1.addCylinder(builderState, va, vb, 0.5, cylinderProps);
        }
    }
    return mesh_builder_1.MeshBuilder.getMesh(builderState);
}
exports.createLinkCylinderMesh = createLinkCylinderMesh;
/**
 * Each edge is included twice to allow for coloring/picking
 * the half closer to the first vertex, i.e. vertex a.
 */
function createLinkLines(ctx, linkBuilder, props, lines) {
    var linkCount = linkBuilder.linkCount, referencePosition = linkBuilder.referencePosition, position = linkBuilder.position, style = linkBuilder.style, ignore = linkBuilder.ignore;
    if (!linkCount)
        return lines_1.Lines.createEmpty(lines);
    var linkScale = props.linkScale, linkSpacing = props.linkSpacing;
    var linesCountEstimate = linkCount * 2;
    var builder = lines_builder_1.LinesBuilder.create(linesCountEstimate, linesCountEstimate / 4, lines);
    var va = linear_algebra_1.Vec3();
    var vb = linear_algebra_1.Vec3();
    var vShift = linear_algebra_1.Vec3();
    for (var edgeIndex = 0, _eI = linkCount; edgeIndex < _eI; ++edgeIndex) {
        if (ignore && ignore(edgeIndex))
            continue;
        position(va, vb, edgeIndex);
        v3scale(vb, v3add(vb, va, vb), 0.5);
        var linkStyle = style ? style(edgeIndex) : 0 /* Solid */;
        if (linkStyle === 0 /* Solid */) {
            builder.add(va[0], va[1], va[2], vb[0], vb[1], vb[2], edgeIndex);
        }
        else if (linkStyle === 1 /* Dashed */) {
            builder.addFixedCountDashes(va, vb, 7, edgeIndex);
        }
        else if (linkStyle === 2 /* Double */ || linkStyle === 3 /* Triple */) {
            var order = 2 /* Double */ ? 2 : 3;
            var multiRadius = 1 * (linkScale / (0.5 * order));
            var absOffset = (1 - multiRadius) * linkSpacing;
            calculateShiftDir(vShift, va, vb, referencePosition ? referencePosition(edgeIndex) : null);
            v3setMagnitude(vShift, vShift, absOffset);
            if (order === 3)
                builder.add(va[0], va[1], va[2], vb[0], vb[1], vb[2], edgeIndex);
            builder.add(va[0] + vShift[0], va[1] + vShift[1], va[2] + vShift[2], vb[0] + vShift[0], vb[1] + vShift[1], vb[2] + vShift[2], edgeIndex);
            builder.add(va[0] - vShift[0], va[1] - vShift[1], va[2] - vShift[2], vb[0] - vShift[0], vb[1] - vShift[1], vb[2] - vShift[2], edgeIndex);
        }
        else if (linkStyle === 4 /* Disk */) {
            v3scale(tmpV12, v3sub(tmpV12, vb, va), 0.475);
            v3add(va, va, tmpV12);
            v3sub(vb, vb, tmpV12);
            // TODO what to do here? Line as disk doesn't work well.
            builder.add(va[0], va[1], va[2], vb[0], vb[1], vb[2], edgeIndex);
        }
    }
    return builder.getLines();
}
exports.createLinkLines = createLinkLines;
//# sourceMappingURL=link.js.map