"use strict";
/**
 * Copyright (c) 2018-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Visual = void 0;
var loci_1 = require("../mol-model/loci");
var marker_action_1 = require("../mol-util/marker-action");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var transform_data_1 = require("../mol-geo/geometry/transform-data");
var util_1 = require("../mol-gl/renderable/util");
var mol_util_1 = require("../mol-util");
var overpaint_data_1 = require("../mol-geo/geometry/overpaint-data");
var int_1 = require("../mol-data/int");
var transparency_data_1 = require("../mol-geo/geometry/transparency-data");
var clipping_data_1 = require("../mol-geo/geometry/clipping-data");
var Visual;
(function (Visual) {
    function setVisibility(renderObject, visible) {
        if (renderObject)
            renderObject.state.visible = visible;
    }
    Visual.setVisibility = setVisibility;
    function setAlphaFactor(renderObject, alphaFactor) {
        if (renderObject)
            renderObject.state.alphaFactor = alphaFactor;
    }
    Visual.setAlphaFactor = setAlphaFactor;
    function setPickable(renderObject, pickable) {
        if (renderObject)
            renderObject.state.pickable = pickable;
    }
    Visual.setPickable = setPickable;
    function setColorOnly(renderObject, colorOnly) {
        if (renderObject)
            renderObject.state.colorOnly = colorOnly;
    }
    Visual.setColorOnly = setColorOnly;
    function mark(renderObject, loci, action, lociApply) {
        if (!renderObject)
            return false;
        var _a = renderObject.values, tMarker = _a.tMarker, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        var array = tMarker.ref.value.array;
        var changed = false;
        if (loci_1.isEveryLoci(loci)) {
            changed = marker_action_1.applyMarkerAction(array, int_1.Interval.ofLength(count), action);
        }
        else if (!loci_1.isEmptyLoci(loci)) {
            changed = lociApply(loci, function (interval) { return marker_action_1.applyMarkerAction(array, interval, action); }, true);
        }
        if (changed)
            mol_util_1.ValueCell.update(tMarker, tMarker.ref.value);
        return changed;
    }
    Visual.mark = mark;
    function setOverpaint(renderObject, overpaint, lociApply, clear) {
        if (!renderObject)
            return;
        var _a = renderObject.values, tOverpaint = _a.tOverpaint, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        // ensure texture has right size
        overpaint_data_1.createOverpaint(overpaint.layers.length ? count : 0, renderObject.values);
        var array = tOverpaint.ref.value.array;
        // clear all if requested
        if (clear)
            overpaint_data_1.clearOverpaint(array, 0, count);
        var _loop_1 = function (i, il) {
            var _a = overpaint.layers[i], loci = _a.loci, color = _a.color, clear_1 = _a.clear;
            var apply = function (interval) {
                var start = int_1.Interval.start(interval);
                var end = int_1.Interval.end(interval);
                return clear_1
                    ? overpaint_data_1.clearOverpaint(array, start, end)
                    : overpaint_data_1.applyOverpaintColor(array, start, end, color);
            };
            lociApply(loci, apply, false);
        };
        for (var i = 0, il = overpaint.layers.length; i < il; ++i) {
            _loop_1(i, il);
        }
        mol_util_1.ValueCell.update(tOverpaint, tOverpaint.ref.value);
    }
    Visual.setOverpaint = setOverpaint;
    function setTransparency(renderObject, transparency, lociApply, clear) {
        if (!renderObject)
            return;
        var _a = renderObject.values, tTransparency = _a.tTransparency, transparencyAverage = _a.transparencyAverage, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        // ensure texture has right size and variant
        transparency_data_1.createTransparency(transparency.layers.length ? count : 0, renderObject.values);
        var array = tTransparency.ref.value.array;
        // clear if requested
        if (clear)
            transparency_data_1.clearTransparency(array, 0, count);
        var _loop_2 = function (i, il) {
            var _a = transparency.layers[i], loci = _a.loci, value = _a.value;
            var apply = function (interval) {
                var start = int_1.Interval.start(interval);
                var end = int_1.Interval.end(interval);
                return transparency_data_1.applyTransparencyValue(array, start, end, value);
            };
            lociApply(loci, apply, false);
        };
        for (var i = 0, il = transparency.layers.length; i < il; ++i) {
            _loop_2(i, il);
        }
        mol_util_1.ValueCell.update(tTransparency, tTransparency.ref.value);
        mol_util_1.ValueCell.updateIfChanged(transparencyAverage, transparency_data_1.getTransparencyAverage(array, count));
    }
    Visual.setTransparency = setTransparency;
    function setClipping(renderObject, clipping, lociApply, clear) {
        if (!renderObject)
            return;
        var _a = renderObject.values, tClipping = _a.tClipping, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        // ensure texture has right size
        clipping_data_1.createClipping(clipping.layers.length ? count : 0, renderObject.values);
        var array = tClipping.ref.value.array;
        // clear if requested
        if (clear)
            clipping_data_1.clearClipping(array, 0, count);
        var _loop_3 = function (i, il) {
            var _a = clipping.layers[i], loci = _a.loci, groups = _a.groups;
            var apply = function (interval) {
                var start = int_1.Interval.start(interval);
                var end = int_1.Interval.end(interval);
                return clipping_data_1.applyClippingGroups(array, start, end, groups);
            };
            lociApply(loci, apply, false);
        };
        for (var i = 0, il = clipping.layers.length; i < il; ++i) {
            _loop_3(i, il);
        }
        mol_util_1.ValueCell.update(tClipping, tClipping.ref.value);
    }
    Visual.setClipping = setClipping;
    function setTransform(renderObject, transform, instanceTransforms) {
        if (!renderObject || (!transform && !instanceTransforms))
            return;
        var values = renderObject.values;
        if (transform) {
            linear_algebra_1.Mat4.copy(values.matrix.ref.value, transform);
            mol_util_1.ValueCell.update(values.matrix, values.matrix.ref.value);
        }
        if (instanceTransforms) {
            values.extraTransform.ref.value.set(instanceTransforms);
            mol_util_1.ValueCell.update(values.extraTransform, values.extraTransform.ref.value);
        }
        else if (instanceTransforms === null) {
            transform_data_1.fillIdentityTransform(values.extraTransform.ref.value, values.instanceCount.ref.value);
            mol_util_1.ValueCell.update(values.extraTransform, values.extraTransform.ref.value);
        }
        transform_data_1.updateTransformData(values);
        var boundingSphere = util_1.calculateTransformBoundingSphere(values.invariantBoundingSphere.ref.value, values.aTransform.ref.value, values.instanceCount.ref.value);
        mol_util_1.ValueCell.update(values.boundingSphere, boundingSphere);
    }
    Visual.setTransform = setTransform;
})(Visual || (Visual = {}));
exports.Visual = Visual;
//# sourceMappingURL=visual.js.map