"use strict";
/**
 * Copyright (c) 2018-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DirectVolumeRepresentationProvider = exports.DirectVolumeRepresentation = exports.DirectVolumeVisual = exports.getDirectVolumeParams = exports.DirectVolumeParams = exports.eachDirectVolume = exports.getDirectVolumeLoci = exports.createDirectVolume = exports.createDirectVolume3d = exports.createDirectVolume2d = void 0;
var tslib_1 = require("tslib");
var param_definition_1 = require("../../mol-util/param-definition");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var geometry_1 = require("../../mol-math/geometry");
var volume_1 = require("../../mol-model/volume");
var direct_volume_1 = require("../../mol-geo/geometry/direct-volume/direct-volume");
var representation_1 = require("./representation");
var location_iterator_1 = require("../../mol-geo/util/location-iterator");
var location_1 = require("../../mol-model/location");
var int_1 = require("../../mol-data/int");
var loci_1 = require("../../mol-model/loci");
var util_1 = require("./util");
// avoiding namespace lookup improved performance in Chrome (Aug 2020)
var v3set = linear_algebra_1.Vec3.set;
var v3normalize = linear_algebra_1.Vec3.normalize;
var v3sub = linear_algebra_1.Vec3.sub;
var v3addScalar = linear_algebra_1.Vec3.addScalar;
var v3scale = linear_algebra_1.Vec3.scale;
var v3toArray = linear_algebra_1.Vec3.toArray;
function getBoundingBox(gridDimension, transform) {
    var bbox = geometry_1.Box3D();
    geometry_1.Box3D.add(bbox, gridDimension);
    geometry_1.Box3D.transform(bbox, bbox, transform);
    return bbox;
}
// 2d volume texture
function getVolumeTexture2dLayout(dim, maxTextureSize) {
    var width = 0;
    var height = dim[1];
    var rows = 1;
    var columns = dim[0];
    if (maxTextureSize < dim[0] * dim[2]) {
        columns = Math.floor(maxTextureSize / dim[0]);
        rows = Math.ceil(dim[2] / columns);
        width = columns * dim[0];
        height *= rows;
    }
    else {
        width = dim[0] * dim[2];
    }
    return { width: width, height: height, columns: columns, rows: rows };
}
function createVolumeTexture2d(volume, maxTextureSize) {
    var _a = volume.grid, _b = _a.cells, space = _b.space, data = _b.data, _c = _a.stats, max = _c.max, min = _c.min;
    var dim = space.dimensions;
    var o = space.dataOffset;
    var _d = getVolumeTexture2dLayout(dim, maxTextureSize), width = _d.width, height = _d.height;
    var array = new Uint8Array(width * height * 4);
    var textureImage = { array: array, width: width, height: height };
    var diff = max - min;
    var xn = dim[0], yn = dim[1], zn = dim[2];
    var n0 = linear_algebra_1.Vec3();
    var n1 = linear_algebra_1.Vec3();
    var xn1 = xn - 1;
    var yn1 = yn - 1;
    var zn1 = zn - 1;
    for (var z = 0; z < zn; ++z) {
        for (var y = 0; y < yn; ++y) {
            for (var x = 0; x < xn; ++x) {
                var column = Math.floor(((z * xn) % width) / xn);
                var row = Math.floor((z * xn) / width);
                var px = column * xn + x;
                var index = 4 * ((row * yn * width) + (y * width) + px);
                var offset = o(x, y, z);
                v3set(n0, data[o(Math.max(0, x - 1), y, z)], data[o(x, Math.max(0, y - 1), z)], data[o(x, y, Math.max(0, z - 1))]);
                v3set(n1, data[o(Math.min(xn1, x + 1), y, z)], data[o(x, Math.min(yn1, y + 1), z)], data[o(x, y, Math.min(zn1, z + 1))]);
                v3normalize(n0, v3sub(n0, n0, n1));
                v3addScalar(n0, v3scale(n0, n0, 0.5), 0.5);
                v3toArray(v3scale(n0, n0, 255), array, index);
                array[index + 3] = ((data[offset] - min) / diff) * 255;
            }
        }
    }
    return textureImage;
}
function createDirectVolume2d(ctx, webgl, volume, directVolume) {
    var gridDimension = volume.grid.cells.space.dimensions;
    var textureImage = createVolumeTexture2d(volume, webgl.maxTextureSize);
    // debugTexture(createImageData(textureImage.array, textureImage.width, textureImage.height), 1/3)
    var transform = volume_1.Grid.getGridToCartesianTransform(volume.grid);
    var bbox = getBoundingBox(gridDimension, transform);
    // TODO: handle disposal
    var texture = directVolume ? directVolume.gridTexture.ref.value : webgl.resources.texture('image-uint8', 'rgba', 'ubyte', 'linear');
    texture.load(textureImage);
    var _a = getUnitToCartn(volume.grid), unitToCartn = _a.unitToCartn, cellDim = _a.cellDim;
    return direct_volume_1.DirectVolume.create(bbox, gridDimension, transform, unitToCartn, cellDim, texture, volume.grid.stats, false, directVolume);
}
exports.createDirectVolume2d = createDirectVolume2d;
// 3d volume texture
function createVolumeTexture3d(volume) {
    var _a = volume.grid, _b = _a.cells, space = _b.space, data = _b.data, _c = _a.stats, max = _c.max, min = _c.min;
    var _d = space.dimensions, width = _d[0], height = _d[1], depth = _d[2];
    var o = space.dataOffset;
    var array = new Uint8Array(width * height * depth * 4);
    var textureVolume = { array: array, width: width, height: height, depth: depth };
    var diff = max - min;
    var n0 = linear_algebra_1.Vec3();
    var n1 = linear_algebra_1.Vec3();
    var width1 = width - 1;
    var height1 = height - 1;
    var depth1 = depth - 1;
    var i = 0;
    for (var z = 0; z < depth; ++z) {
        for (var y = 0; y < height; ++y) {
            for (var x = 0; x < width; ++x) {
                var offset = o(x, y, z);
                v3set(n0, data[o(Math.max(0, x - 1), y, z)], data[o(x, Math.max(0, y - 1), z)], data[o(x, y, Math.max(0, z - 1))]);
                v3set(n1, data[o(Math.min(width1, x + 1), y, z)], data[o(x, Math.min(height1, y + 1), z)], data[o(x, y, Math.min(depth1, z + 1))]);
                v3normalize(n0, v3sub(n0, n0, n1));
                v3addScalar(n0, v3scale(n0, n0, 0.5), 0.5);
                v3toArray(v3scale(n0, n0, 255), array, i);
                array[i + 3] = ((data[offset] - min) / diff) * 255;
                i += 4;
            }
        }
    }
    return textureVolume;
}
function getUnitToCartn(grid) {
    if (grid.transform.kind === 'matrix') {
        return {
            unitToCartn: linear_algebra_1.Mat4.mul(linear_algebra_1.Mat4(), grid.transform.matrix, linear_algebra_1.Mat4.fromScaling(linear_algebra_1.Mat4(), grid.cells.space.dimensions)),
            cellDim: linear_algebra_1.Mat4.getScaling(linear_algebra_1.Vec3(), grid.transform.matrix)
        };
    }
    var box = grid.transform.fractionalBox;
    var size = geometry_1.Box3D.size(linear_algebra_1.Vec3(), box);
    return {
        unitToCartn: linear_algebra_1.Mat4.mul3(linear_algebra_1.Mat4(), grid.transform.cell.fromFractional, linear_algebra_1.Mat4.fromTranslation(linear_algebra_1.Mat4(), box.min), linear_algebra_1.Mat4.fromScaling(linear_algebra_1.Mat4(), size)),
        cellDim: linear_algebra_1.Vec3.div(linear_algebra_1.Vec3(), grid.transform.cell.size, grid.cells.space.dimensions)
    };
}
function createDirectVolume3d(ctx, webgl, volume, directVolume) {
    var gridDimension = volume.grid.cells.space.dimensions;
    var textureVolume = createVolumeTexture3d(volume);
    var transform = volume_1.Grid.getGridToCartesianTransform(volume.grid);
    var bbox = getBoundingBox(gridDimension, transform);
    // TODO: handle disposal
    var texture = directVolume ? directVolume.gridTexture.ref.value : webgl.resources.texture('volume-uint8', 'rgba', 'ubyte', 'linear');
    texture.load(textureVolume);
    var _a = getUnitToCartn(volume.grid), unitToCartn = _a.unitToCartn, cellDim = _a.cellDim;
    return direct_volume_1.DirectVolume.create(bbox, gridDimension, transform, unitToCartn, cellDim, texture, volume.grid.stats, false, directVolume);
}
exports.createDirectVolume3d = createDirectVolume3d;
//
function createDirectVolume(ctx, volume, theme, props, directVolume) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var runtime, webgl;
        return tslib_1.__generator(this, function (_a) {
            runtime = ctx.runtime, webgl = ctx.webgl;
            if (webgl === undefined)
                throw new Error('DirectVolumeVisual requires `webgl` in props');
            return [2 /*return*/, webgl.isWebGL2 ?
                    createDirectVolume3d(runtime, webgl, volume, directVolume) :
                    createDirectVolume2d(runtime, webgl, volume, directVolume)];
        });
    });
}
exports.createDirectVolume = createDirectVolume;
function getLoci(volume, props) {
    return props.renderMode.name === 'isosurface'
        ? volume_1.Volume.Isosurface.Loci(volume, props.renderMode.params.isoValue)
        : volume_1.Volume.Loci(volume);
}
function getDirectVolumeLoci(pickingId, volume, props, id) {
    var objectId = pickingId.objectId, groupId = pickingId.groupId;
    if (id === objectId) {
        return volume_1.Volume.Cell.Loci(volume, int_1.Interval.ofSingleton(groupId));
    }
    return loci_1.EmptyLoci;
}
exports.getDirectVolumeLoci = getDirectVolumeLoci;
function eachDirectVolume(loci, volume, props, apply) {
    var isoValue = props.renderMode.name === 'isosurface'
        ? props.renderMode.params.isoValue : undefined;
    return util_1.eachVolumeLoci(loci, volume, isoValue, apply);
}
exports.eachDirectVolume = eachDirectVolume;
//
exports.DirectVolumeParams = tslib_1.__assign(tslib_1.__assign({}, direct_volume_1.DirectVolume.Params), { quality: tslib_1.__assign(tslib_1.__assign({}, direct_volume_1.DirectVolume.Params.quality), { isEssential: false }) });
function getDirectVolumeParams(ctx, volume) {
    var p = param_definition_1.ParamDefinition.clone(exports.DirectVolumeParams);
    p.renderMode = direct_volume_1.DirectVolume.createRenderModeParam(volume.grid.stats);
    return p;
}
exports.getDirectVolumeParams = getDirectVolumeParams;
function DirectVolumeVisual(materialId) {
    return representation_1.VolumeVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.DirectVolumeParams),
        createGeometry: createDirectVolume,
        createLocationIterator: function (volume) { return location_iterator_1.LocationIterator(volume.grid.cells.data.length, 1, 1, function () { return location_1.NullLocation; }); },
        getLoci: getDirectVolumeLoci,
        eachLocation: eachDirectVolume,
        setUpdateState: function (state, volume, newProps, currentProps) {
        },
        geometryUtils: direct_volume_1.DirectVolume.Utils
    }, materialId);
}
exports.DirectVolumeVisual = DirectVolumeVisual;
function DirectVolumeRepresentation(ctx, getParams) {
    return representation_1.VolumeRepresentation('Direct Volume', ctx, getParams, DirectVolumeVisual, getLoci);
}
exports.DirectVolumeRepresentation = DirectVolumeRepresentation;
exports.DirectVolumeRepresentationProvider = representation_1.VolumeRepresentationProvider({
    name: 'direct-volume',
    label: 'Direct Volume',
    description: 'Direct rendering of volumetric data.',
    factory: DirectVolumeRepresentation,
    getParams: getDirectVolumeParams,
    defaultValues: param_definition_1.ParamDefinition.getDefaultValues(exports.DirectVolumeParams),
    defaultColorTheme: { name: 'uniform' },
    defaultSizeTheme: { name: 'uniform' },
    isApplicable: function (volume) { return !volume_1.Volume.isEmpty(volume); }
});
//# sourceMappingURL=direct-volume.js.map