/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { __assign } from "tslib";
import { QuadSchema, QuadValues } from '../../mol-gl/compute/util';
import { TextureSpec, UniformSpec, DefineSpec } from '../../mol-gl/renderable/schema';
import { ShaderCode } from '../../mol-gl/shader-code';
import { ValueCell } from '../../mol-util';
import { createComputeRenderItem } from '../../mol-gl/webgl/render-item';
import { createComputeRenderable } from '../../mol-gl/renderable';
import { Vec2, Vec3 } from '../../mol-math/linear-algebra';
import { ParamDefinition as PD } from '../../mol-util/param-definition';
import quad_vert from '../../mol-gl/shader/quad.vert';
import postprocessing_frag from '../../mol-gl/shader/postprocessing.frag';
import fxaa_frag from '../../mol-gl/shader/fxaa.frag';
import { StereoCamera } from '../camera/stereo';
var PostprocessingSchema = __assign(__assign({}, QuadSchema), { tColor: TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), tPackedDepth: TextureSpec('texture', 'depth', 'ushort', 'nearest'), uTexSize: UniformSpec('v2'), dOrthographic: DefineSpec('number'), uNear: UniformSpec('f'), uFar: UniformSpec('f'), uFogNear: UniformSpec('f'), uFogFar: UniformSpec('f'), uFogColor: UniformSpec('v3'), dOcclusionEnable: DefineSpec('boolean'), dOcclusionKernelSize: DefineSpec('number'), uOcclusionBias: UniformSpec('f'), uOcclusionRadius: UniformSpec('f'), dOutlineEnable: DefineSpec('boolean'), uOutlineScale: UniformSpec('f'), uOutlineThreshold: UniformSpec('f') });
var PostprocessingShaderCode = ShaderCode('postprocessing', quad_vert, postprocessing_frag);
function getPostprocessingRenderable(ctx, colorTexture, depthTexture) {
    var width = colorTexture.getWidth();
    var height = colorTexture.getHeight();
    var values = __assign(__assign({}, QuadValues), { tColor: ValueCell.create(colorTexture), tPackedDepth: ValueCell.create(depthTexture), uTexSize: ValueCell.create(Vec2.create(width, height)), dOrthographic: ValueCell.create(0), uNear: ValueCell.create(1), uFar: ValueCell.create(10000), uFogNear: ValueCell.create(10000), uFogFar: ValueCell.create(10000), uFogColor: ValueCell.create(Vec3.create(1, 1, 1)), dOcclusionEnable: ValueCell.create(false), dOcclusionKernelSize: ValueCell.create(4), uOcclusionBias: ValueCell.create(0.5), uOcclusionRadius: ValueCell.create(64), dOutlineEnable: ValueCell.create(false), uOutlineScale: ValueCell.create(1 * ctx.pixelRatio), uOutlineThreshold: ValueCell.create(0.8) });
    var schema = __assign({}, PostprocessingSchema);
    var renderItem = createComputeRenderItem(ctx, 'triangles', PostprocessingShaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
export var PostprocessingParams = {
    occlusion: PD.MappedStatic('off', {
        on: PD.Group({
            kernelSize: PD.Numeric(4, { min: 1, max: 32, step: 1 }),
            bias: PD.Numeric(0.5, { min: 0, max: 1, step: 0.01 }),
            radius: PD.Numeric(64, { min: 0, max: 256, step: 1 }),
        }),
        off: PD.Group({})
    }, { cycle: true, description: 'Darken occluded crevices with the ambient occlusion effect' }),
    outline: PD.MappedStatic('off', {
        on: PD.Group({
            scale: PD.Numeric(1, { min: 0, max: 10, step: 1 }),
            threshold: PD.Numeric(0.8, { min: 0, max: 5, step: 0.01 }),
        }),
        off: PD.Group({})
    }, { cycle: true, description: 'Draw outline around 3D objects' }),
    antialiasing: PD.MappedStatic('on', {
        on: PD.Group({
            edgeThresholdMin: PD.Numeric(0.0312, { min: 0.0312, max: 0.0833, step: 0.0001 }, { description: 'Trims the algorithm from processing darks.' }),
            edgeThresholdMax: PD.Numeric(0.063, { min: 0.063, max: 0.333, step: 0.001 }, { description: 'The minimum amount of local contrast required to apply algorithm.' }),
            iterations: PD.Numeric(12, { min: 0, max: 32, step: 1 }, { description: 'Number of edge exploration steps.' }),
            subpixelQuality: PD.Numeric(1.00, { min: 0.00, max: 1.00, step: 0.01 }, { description: 'Choose the amount of sub-pixel aliasing removal.' }),
        }),
        off: PD.Group({})
    }, { cycle: true, description: 'Fast Approximate Anti-Aliasing (FXAA)' }),
};
var PostprocessingPass = /** @class */ (function () {
    function PostprocessingPass(webgl, drawPass) {
        this.webgl = webgl;
        this.drawPass = drawPass;
        var colorTarget = drawPass.colorTarget, depthTexture = drawPass.depthTexture;
        var width = colorTarget.getWidth();
        var height = colorTarget.getHeight();
        this.target = webgl.createRenderTarget(width, height, false);
        this.tmpTarget = webgl.createRenderTarget(width, height, false, 'uint8', 'linear');
        this.renderable = getPostprocessingRenderable(webgl, colorTarget.texture, depthTexture);
        this.fxaa = getFxaaRenderable(webgl, this.tmpTarget.texture);
    }
    PostprocessingPass.isEnabled = function (props) {
        return props.occlusion.name === 'on' || props.outline.name === 'on' || props.antialiasing.name === 'on';
    };
    PostprocessingPass.prototype.syncSize = function () {
        var width = this.drawPass.colorTarget.getWidth();
        var height = this.drawPass.colorTarget.getHeight();
        var _a = this.renderable.values.uTexSize.ref.value, w = _a[0], h = _a[1];
        if (width !== w || height !== h) {
            this.target.setSize(width, height);
            this.tmpTarget.setSize(width, height);
            ValueCell.update(this.renderable.values.uTexSize, Vec2.set(this.renderable.values.uTexSize.ref.value, width, height));
            ValueCell.update(this.fxaa.values.uTexSizeInv, Vec2.set(this.fxaa.values.uTexSizeInv.ref.value, 1 / width, 1 / height));
        }
    };
    PostprocessingPass.prototype.updateState = function (camera) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        state.enable(gl.SCISSOR_TEST);
        state.disable(gl.BLEND);
        state.disable(gl.DEPTH_TEST);
        state.depthMask(false);
        var _b = camera.viewport, x = _b.x, y = _b.y, width = _b.width, height = _b.height;
        gl.viewport(x, y, width, height);
        gl.scissor(x, y, width, height);
        state.clearColor(0, 0, 0, 1);
        gl.clear(gl.COLOR_BUFFER_BIT);
    };
    PostprocessingPass.prototype._renderPostprocessing = function (camera, toDrawingBuffer, props) {
        var values = this.renderable.values;
        ValueCell.updateIfChanged(values.uFar, camera.far);
        ValueCell.updateIfChanged(values.uNear, camera.near);
        ValueCell.updateIfChanged(values.uFogFar, camera.fogFar);
        ValueCell.updateIfChanged(values.uFogNear, camera.fogNear);
        var needsUpdate = false;
        var orthographic = camera.state.mode === 'orthographic' ? 1 : 0;
        if (values.dOrthographic.ref.value !== orthographic)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dOrthographic, orthographic);
        var occlusion = props.occlusion.name === 'on';
        if (values.dOcclusionEnable.ref.value !== occlusion)
            needsUpdate = true;
        ValueCell.updateIfChanged(this.renderable.values.dOcclusionEnable, occlusion);
        if (props.occlusion.name === 'on') {
            var kernelSize = props.occlusion.params.kernelSize;
            if (values.dOcclusionKernelSize.ref.value !== kernelSize)
                needsUpdate = true;
            ValueCell.updateIfChanged(values.dOcclusionKernelSize, kernelSize);
            ValueCell.updateIfChanged(values.uOcclusionBias, props.occlusion.params.bias);
            ValueCell.updateIfChanged(values.uOcclusionRadius, props.occlusion.params.radius);
        }
        var outline = props.outline.name === 'on';
        if (values.dOutlineEnable.ref.value !== outline)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dOutlineEnable, outline);
        if (props.outline.name === 'on') {
            ValueCell.updateIfChanged(values.uOutlineScale, props.outline.params.scale * this.webgl.pixelRatio);
            ValueCell.updateIfChanged(values.uOutlineThreshold, props.outline.params.threshold);
        }
        if (needsUpdate) {
            this.renderable.update();
        }
        if (props.antialiasing.name === 'on') {
            this.tmpTarget.bind();
        }
        else if (toDrawingBuffer) {
            this.webgl.unbindFramebuffer();
        }
        else {
            this.target.bind();
        }
        this.updateState(camera);
        this.renderable.render();
    };
    PostprocessingPass.prototype._renderFxaa = function (camera, toDrawingBuffer, props) {
        if (props.antialiasing.name === 'off')
            return;
        var values = this.fxaa.values;
        var needsUpdate = false;
        var input = (props.occlusion.name === 'on' || props.outline.name === 'on')
            ? this.tmpTarget.texture : this.drawPass.colorTarget.texture;
        if (values.tColor.ref.value !== input) {
            ValueCell.update(this.fxaa.values.tColor, input);
            needsUpdate = true;
        }
        var _a = props.antialiasing.params, edgeThresholdMin = _a.edgeThresholdMin, edgeThresholdMax = _a.edgeThresholdMax, iterations = _a.iterations, subpixelQuality = _a.subpixelQuality;
        if (values.dEdgeThresholdMin.ref.value !== edgeThresholdMin)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dEdgeThresholdMin, edgeThresholdMin);
        if (values.dEdgeThresholdMax.ref.value !== edgeThresholdMax)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dEdgeThresholdMax, edgeThresholdMax);
        if (values.dIterations.ref.value !== iterations)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dIterations, iterations);
        if (values.dSubpixelQuality.ref.value !== subpixelQuality)
            needsUpdate = true;
        ValueCell.updateIfChanged(values.dSubpixelQuality, subpixelQuality);
        if (needsUpdate) {
            this.fxaa.update();
        }
        if (toDrawingBuffer) {
            this.webgl.unbindFramebuffer();
        }
        else {
            this.target.bind();
        }
        this.updateState(camera);
        this.fxaa.render();
    };
    PostprocessingPass.prototype._render = function (camera, toDrawingBuffer, props) {
        if (props.occlusion.name === 'on' || props.outline.name === 'on' || props.antialiasing.name === 'off') {
            this._renderPostprocessing(camera, toDrawingBuffer, props);
        }
        if (props.antialiasing.name === 'on') {
            this._renderFxaa(camera, toDrawingBuffer, props);
        }
    };
    PostprocessingPass.prototype.render = function (camera, toDrawingBuffer, props) {
        if (StereoCamera.is(camera)) {
            this._render(camera.left, toDrawingBuffer, props);
            this._render(camera.right, toDrawingBuffer, props);
        }
        else {
            this._render(camera, toDrawingBuffer, props);
        }
    };
    return PostprocessingPass;
}());
export { PostprocessingPass };
//
var FxaaSchema = __assign(__assign({}, QuadSchema), { tColor: TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), uTexSizeInv: UniformSpec('v2'), dEdgeThresholdMin: DefineSpec('number'), dEdgeThresholdMax: DefineSpec('number'), dIterations: DefineSpec('number'), dSubpixelQuality: DefineSpec('number') });
var FxaaShaderCode = ShaderCode('fxaa', quad_vert, fxaa_frag);
function getFxaaRenderable(ctx, colorTexture) {
    var width = colorTexture.getWidth();
    var height = colorTexture.getHeight();
    var values = __assign(__assign({}, QuadValues), { tColor: ValueCell.create(colorTexture), uTexSizeInv: ValueCell.create(Vec2.create(1 / width, 1 / height)), dEdgeThresholdMin: ValueCell.create(0.0312), dEdgeThresholdMax: ValueCell.create(0.125), dIterations: ValueCell.create(12), dSubpixelQuality: ValueCell.create(0.75) });
    var schema = __assign({}, FxaaSchema);
    var renderItem = createComputeRenderItem(ctx, 'triangles', FxaaShaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
//# sourceMappingURL=postprocessing.js.map